#ifndef MENU_OBJECT_H
#define MENU_OBJECT_H

#include <string>
#include <list>
#include <cassert>

#include "defs/defs.h"
#include "math/mat.h"
#include "gc/pointer.h"
#include "gc/object.h"

// ***************************** MENU ***************************** 

namespace pwn {

class MenuSystem;

//this class is a bit of a mystery
//why not just use the mouse global object?
class MouseEvent {
public:
	MouseEvent()
	:	lpress(0),
		rpress(0),
		mpress(0),
		ldown(0),
		rdown(0),
		mdown(0),
		lrelease(0),
		rrelease(0),
		mrelease(0) {}
	//when using sets of unsigneds the compiler will round them up to the nearest int
	unsigned lpress : 1;
	unsigned rpress : 1;
	unsigned mpress : 1;
	unsigned ldown : 1;
	unsigned rdown : 1;
	unsigned mdown : 1;
	unsigned lrelease : 1;
	unsigned rrelease : 1;
	unsigned mrelease : 1;
};

enum eventResponse_t {
	//used by preMouseEvent & mouseEvent
	RESPONSE_CONTINUE,		//everything is a ok
	RESPONSE_STOP,			//don't process our children
	//only used by preMouseEvent.  mouseEvent will treat as a STOP. (only used there because mouseEvent can't handle it - because by then the children are already being processed)
	RESPONSE_HALT,			//hald in procedure
};

class Object;
class ResizeListener {
public:
	virtual void callback_onResize(Object *o) = 0;
};

class Object : public gc::Object {
protected:
	friend class MenuSystem;

	//system
	bool active;
	gc::Pointer<Object> parent;
	std::list<gc::Pointer<Object> > kids;

	bool visible;				//whether to display and handle input or not
	
	//mouse bubbling progression - parent-to-child
	gc::Pointer<Object> selectedChild;
	
	//keyboard bubbling progression - parent-to-child
	gc::Pointer<Object> focusChild;
	
	vec2f pos;					//for quad geom / input area
	vec2f size;					//for occlusion of children / input event association
	vec2f scale;				//for scaling children
	unsigned int backgroundColor;
	int textureID;				//TODO - make this "backgroundImage"
	unsigned int fontColor;				//text =) currently writes black background - i could use an add blend to fix this
	unsigned int borderColor;			//around the obj
	vec2f fontSize;				//scale of the font
	
	vec2f childofs;				//mainly for scrolling
	int topmostPriority;		//in the tradition windows way - with a thread-system twist.  - = lower, + = higher
	
	//whether the menu is allowed to get or lose focus
	bool allowFocus;

	//most menus have text in them ... why not all?!
	std::string text;
	
	//whether objects beneath this can still receive input
	bool occludesInput;

	std::list<ResizeListener* > resizeListeners;
	
	//nine patch stuff
	bool useNinePatch;
	float ninePatchUVBorder;
	float ninePatchBorder;

public:

	virtual void addResizeListener(ResizeListener *l);
	virtual void removeResizeListener(ResizeListener *l);

	//now, as I add these, I cannot put their methods in the .h file
	//the methods cannot be inline because they reference the 'engine' api,
	//which cannot be in the header
	//
	//for static linking this will link both the game and engine methods to the game code
	//but for dynamic linking, the engine will not be able to see the game's method implementations
	//therefore the engine will have to implement its own
	//
	//my suggestion to this problem:
	//either try as hard as can be to not use any menu methods engine-side 
	//	(only possible by not instanciating any menus whatsoever)
	//or, for these base methods, only provide direct callbacks through the associated api methods
	//implement this on both sides, engine and game, and things should go smooth

	Object(const vec2f &_pos = vec2f(0,0), const vec2f &_size = vec2f(1,1));
protected:
	virtual ~Object();	//only I can delete them!
public:

	void addChild(Object *child);
	void unlinkFromParent();
	int indexOfChild(Object * m);
	
	static Object * getRoot();

	//and here's the rest of the story!

	bool isActive() { return active; }
	
	//tell me about const list ... you can't add to it, right?  but you still get to access its members as non-const, right?
	std::list<gc::Pointer<Object> > &getKids() { return kids; }
	
	Object *getParent() { return parent.ptr(); }
	
	//input system callbacks
	virtual eventResponse_t preMouseEvent(MouseEvent event, const vec2f &v);
	virtual eventResponse_t mouseEvent(MouseEvent event, const vec2f &v);
	
	//left-click-on-me responses
	virtual void onFocus() {}
	virtual void onBlur() {}
	
	//only called - during the menu input system - if this menu has focus
	//returns whether we needed the key (whether to commendere)
	virtual bool doKeyPress(int key) { return false; }

	//the user calls these if they want to change the focus of the menu
	virtual void setFocus();
	virtual void loseFocus();
	
	//display system
	virtual bool display();	//return whether or not we want to also draw children
	virtual void postDisplay();	//runs after the children are displayed

	virtual vec2f getRelPos(Object *rel);
	vec2f getAbsPos();

	virtual mat23f getRelBasis(Object *rel) const;
	virtual mat23f getRelInvBasis(Object *rel) const;
	virtual mat23f getTransform() const;
	virtual mat23f getInvTransform() const;
	virtual mat23f getChildTransform() const;
	virtual mat23f getInvChildTransform() const;

	//variable interfaces
	virtual vec2f getPos() const { return pos; }
	virtual void setPos(const vec2f &_pos) { pos = _pos; }
	
	virtual vec2f getSize() const { return size; }
	virtual void setSize(const vec2f &_size);
	
	virtual vec2f getScale() const { return scale; }
	virtual void setScale(const vec2f &_scale) { scale = _scale; }
	
	virtual bool getVisible() { return visible; }
	virtual void setVisible(bool _visible) { visible = _visible; }
	
	virtual unsigned int getBackgroundColor() const { return backgroundColor; }
	virtual void setBackgroundColor(unsigned int backgroundColor_) { backgroundColor = backgroundColor_; }
	
	virtual unsigned int getFontColor() const { return fontColor; }
	virtual void setFontColor(unsigned int _fontColor) { fontColor = _fontColor; }

	virtual unsigned int getBorderColor() const { return borderColor; }
	virtual void setBorderColor(unsigned int _borderColor) { borderColor = _borderColor; }
	
	virtual vec2f getFontSize() const { return fontSize; }
	virtual void setFontSize(const vec2f &_fontSize) { fontSize = _fontSize; }
	
	virtual int getTextureID() const { return textureID; }
	virtual void setTextureID(int _textureID) { textureID = _textureID; }
	
	virtual int getTopmostPriority() const { return topmostPriority; }
	virtual void setTopmostPriority(int _topmostPriority) { topmostPriority = _topmostPriority; }
	
	virtual bool getAllowFocus() const { return allowFocus; }
	virtual void setAllowFocus(bool _allowFocus) { allowFocus = _allowFocus; }
	
	virtual const std::string &getText() const { return text; }
	virtual void setText(const std::string &_text) { text = _text; }
	
	virtual bool getUseNinePatch() const { return useNinePatch; }
	virtual void setUseNinePatch(bool useNinePatch_) { useNinePatch = useNinePatch_; }
	
	virtual float getNinePatchUVBorder() const { return ninePatchUVBorder; }
	virtual void setNinePatchUVBorder(float ninePatchUVBorder_) { ninePatchUVBorder = ninePatchUVBorder_; }
	
	virtual float getNinePatchBorder() const { return ninePatchBorder; }
	virtual void setNinePatchBorder(float ninePatchBorder_) { ninePatchBorder = ninePatchBorder_; }
	
	virtual bool getOccludesInput() const { return occludesInput; }
	virtual void setOccludesInput(bool occludesInput_) { occludesInput = occludesInput_; }
};

};

#endif	//MENU_OBJECT_H
